#include <Conn_config.h>

#ifdef POLL_FOUND

#include <Conn_engine_poll.h>

/*
 * Variables
 */
static unsigned long	Conn_poll_allocated_slots = 0;
static struct pollfd	*Conn_poll_pfds = NULL;

/*
 * Functions
 */

/*
 * Init the engine
 */
int Conn_poll_init(void)
{
	return 0;
}

int Conn_poll_shutdown(void)
{
	free(Conn_poll_pfds);

	return 0;
}

/*
 * Grow the number of slots needed for poll
 */
int Conn_poll_grow(unsigned int alloc)
{
	struct pollfd *p, *set;
	unsigned int diff, mem;

	Log(10, "%s: Trying to grow pollfds from %d to %d.\n",
		__FUNCTION__,
		Conn_poll_allocated_slots, alloc);

	if (alloc <= Conn_poll_allocated_slots)
		return 0;

	diff = alloc - Conn_poll_allocated_slots;

	mem = alloc * sizeof(struct pollfd);
	p = (struct pollfd *) realloc(Conn_poll_pfds, mem);
	if (p == NULL) {
		snprintf(Conn_error, sizeof(Conn_error),
			"Cannot alloc %u bytes for pollfds.", mem);
		return -1;
	}

	set = p + Conn_poll_allocated_slots;
	memset(set, 0, diff * sizeof(struct pollfd));
	Conn_poll_pfds = p;

	return 0;
}

/*
 * Add a target to the list
 */
int Conn_poll_add_obj(struct Conn *C)
{
	Conn_poll_pfds[C->slot].fd = C->fd;
	Conn_poll_pfds[C->slot].events = C->events;

	return 0;
}

/*
 * Remove a target from the list
 */
int Conn_poll_del_obj(struct Conn *C)
{
	Conn_poll_pfds[C->slot].fd = -1;
	Conn_poll_pfds[C->slot].events = 0;

	return 0;
}

/*
 * Change event mask
 */
int Conn_poll_chg_obj(struct Conn *C)
{
	Conn_poll_pfds[C->slot].events = C->events;

	return 0;
}


/*
 * Calls a callback for fds with activity
 * Returns: -1 on error, 0 nothing to do, n (>0) if some work was done
 * timeout is in 1/1000 seconds increments.
 */
int Conn_poll_poll(const int timeout2, void (*cb)(const unsigned int slot,
	const int revents))
{
	int slot, events;

	again:
	events = poll(&Conn_poll_pfds[0], Conn_no, timeout2);
	if ((events == -1) && (errno == EINTR))
		goto again;

	if (events < 0) {
		snprintf(Conn_error, sizeof(Conn_error),
			"%s: Conn_no=%d [%s]",
			__FUNCTION__, Conn_no,
			strerror(errno));
		return -1;
	}

	gettimeofday(&Conn_now, NULL);

	if (events == 0)
		return 0;

	/* We do revers scan because of moving Conn objects */
	Log(11, "Processing %d event(s)...\n",
		events);
	slot = Conn_no - 1;
	do {
		if (Conn_poll_pfds[slot].revents & POLLNVAL) {
			Log(0, "BUG NVAL!\n");
			exit(1);
		}

		cb(slot, Conn_poll_pfds[slot].revents);

		slot--;
	} while (slot >= 0);

	return events;

}

/*
 * Move a slot over an other.
 */
int Conn_poll_move_slot(const unsigned int dst, const unsigned int src)
{
	struct pollfd tmp;

	tmp = Conn_poll_pfds[dst];
	Conn_poll_pfds[dst] = Conn_poll_pfds[src];
	Conn_poll_pfds[src] = tmp;

	return 0;
}

#endif
